#include "cpu.h"
#include "pic32_c.h"
#include "init.h"



/* Basic memory and pin initialisation. */

void init_memory(void)
{
    /*
    Configure RAM.
    See: http://microchipdeveloper.com/32bit:mx-arch-exceptions-processor-initialization
    */

    uint32_t config = REG(BMXCON);

    /* Set zero wait states for address setup. */

    config &= ~(1 << 6);        /* BMXCON<6> = BMXWSDRM = 0 */

    /* Set bus arbitration mode. */

    config &= ~0b111;
    config |= 0b010;            /* BMXCON<2:0> = BMXARB<2:0> = 2 */

    REG(BMXCON) = config;
}

void init_pins(void)
{
    /* DEVCFG0<2> also needs setting to 0 before the program is run. */

    CLR_REG(CFGCON, 1 << 3);    /* CFGCON<3> = JTAGEN = 0 */
}

void init_outputs(void)
{
    /* Remove analogue features from pins. */

    REG(ANSELA) = 0;
    REG(ANSELB) = 0;

    /* Set pins as outputs. */

    REG(TRISA) = 0;
    REG(TRISB) = 0;

    /* Clear outputs. */

    REG(PORTA) = 0;
    REG(PORTB) = 0;
}



/* Peripheral pin configuration. */

void config_uart(void)
{
    /* NOTE: Configuring UART1 for specific pins. */

    /* Map U1RX to RPB13. */

    REG(U1RXR) = 0b0011;        /* U1RXR<3:0> = 0011 (RPB13) */

    /* Map U1TX to RPB15. */

    REG(RPB15R) = 0b0001;       /* RPB15R<3:0> = 0001 (U1TX) */

    /* Set RPB13 to input. */

    SET_REG(TRISB, 1 << 13);
}

void lock_config(void)
{
    SET_REG(CFGCON, 1 << 13);   /* IOLOCK = 1 */

    /* Lock the configuration again. */

    REG(SYSKEY) = 0x33333333;
}

void unlock_config(void)
{
    /* Unlock the configuration register bits. */

    REG(SYSKEY) = 0;
    REG(SYSKEY) = 0xAA996655;
    REG(SYSKEY) = 0x556699AA;

    CLR_REG(CFGCON, 1 << 13);   /* IOLOCK = 0 */
}



/* Convenience operations. */

void interrupts_on(void)
{
    init_interrupts();
    enable_interrupts();
    handle_error_level();
}



/* DMA configuration. */

void init_dma(void)
{
    /* Disable DMA interrupts. */

    CLR_REG(DMAIEC, 0b1111 << DMAINTBASE); /* DMA3IE...DMA0IE = 0 */

    /* Clear DMA interrupt flags. */

    CLR_REG(DMAIFS, 0b1111 << DMAINTBASE); /* DMA3IF...DMA0IF = 0 */

    /* Enable DMA. */

    SET_REG(DMACON, 1 << 15);
}

/* Initialise the given channel. */

void dma_init(int channel, uint8_t pri)
{
    if ((channel < DCHMIN) || (channel > DCHMAX))
        return;

    /* Initialise a channel. */

    REG(DMA_REG(channel, DCHxCON)) = pri & 0b11;
    REG(DMA_REG(channel, DCHxECON)) = 0;
    REG(DMA_REG(channel, DCHxINT)) = 0;
}

/* Set the channel auto-enable mode. */

void dma_set_auto_enable(int channel, int auto_enable)
{
    (auto_enable ? SET_REG : CLR_REG)(DMA_REG(channel, DCHxCON), 1 << 4);
}

/* Set the channel chaining mode. */

void dma_set_chaining(int channel, enum dma_chain chain)
{
    (chain != dma_chain_none ?
     SET_REG : CLR_REG)(DMA_REG(channel, DCHxCON), 1 << 5);

    (chain == dma_chain_next ?
     SET_REG : CLR_REG)(DMA_REG(channel, DCHxCON), 1 << 8);
}

/* Configure a channel's initiation interrupt. */

void dma_set_interrupt(int channel, uint8_t int_num, int enable)
{
    if ((channel < DCHMIN) || (channel > DCHMAX))
        return;

    /* Allow an interrupt to trigger the transfer. */

    REG(DMA_REG(channel, DCHxECON)) = (int_num << 8) |
                                      ((enable ? 1 : 0) << 4);
}

/* Set a channel's transfer parameters. */

void dma_set_transfer(int channel,
                  uint32_t source_start_address, uint16_t source_size,
                  uint32_t destination_start_address, uint16_t destination_size,
                  uint16_t cell_size)
{
    if ((channel < DCHMIN) || (channel > DCHMAX))
        return;

    REG(DMA_REG(channel, DCHxSSIZ)) = source_size;
    REG(DMA_REG(channel, DCHxSSA)) = source_start_address;
    REG(DMA_REG(channel, DCHxDSIZ)) = destination_size;
    REG(DMA_REG(channel, DCHxDSA)) = destination_start_address;
    REG(DMA_REG(channel, DCHxCSIZ)) = cell_size;
}

/* Configure interrupts caused by the channel. */

void dma_init_interrupt(int channel, uint8_t conditions,
                        uint8_t pri, uint8_t sub)
{
    if ((channel < DCHMIN) || (channel > DCHMAX))
        return;

    /* Disable channel interrupt and clear interrupt flag. */

    CLR_REG(DMAIEC, DMA_INT_FLAGS(channel, 1));
    CLR_REG(DMAIFS, DMA_INT_FLAGS(channel, 1));

    /* Produce an interrupt for the provided conditions. */

    REG(DMA_REG(channel, DCHxINT)) = conditions << 16;

    /* Set interrupt priorities. */

    REG(DMAIPC) = (REG(DMAIPC) &
                   ~(DMA_IPC_PRI(channel, 7, 3))) |
                  DMA_IPC_PRI(channel, pri, sub);

    /* Enable interrupt. */

    SET_REG(DMAIEC, DMA_INT_FLAGS(channel, 1));
}

/* Enable a DMA channel. */

void dma_on(int channel)
{
    if ((channel < DCHMIN) || (channel > DCHMAX))
        return;

    /* Enable channel. */

    SET_REG(DMA_REG(channel, DCHxCON), 1 << 7);
}



/* Output compare configuration. */

void oc_init(int unit, uint8_t mode, int timer)
{
    if ((unit < OCMIN) || (unit > OCMAX))
        return;

    REG(OC_REG(unit, OCxCON)) = (timer == 3 ? (1 << 3) : 0) | (mode & 0b111);
}

/* Set the start value for the pulse. */

void oc_set_pulse(int unit, uint32_t start)
{
    if ((unit < OCMIN) || (unit > OCMAX))
        return;

    REG(OC_REG(unit, OCxR)) = start;
}

/* Set the end value for the pulse. */

void oc_set_pulse_end(int unit, uint32_t end)
{
    if ((unit < OCMIN) || (unit > OCMAX))
        return;

    REG(OC_REG(unit, OCxRS)) = end;
}

/* Configure interrupts caused by the unit. */

void oc_init_interrupt(int unit, uint8_t pri, uint8_t sub)
{
    if ((unit < OCMIN) || (unit > OCMAX))
        return;

    /* Disable interrupt and clear interrupt flag. */

    CLR_REG(OCIEC, OC_INT_FLAGS(unit, OCxIE));
    CLR_REG(OCIFS, OC_INT_FLAGS(unit, OCxIF));

    /* Set interrupt priorities. */

    REG(OC_IPC_REG(unit)) = (REG(OC_IPC_REG(unit)) &
                             ~(OC_IPC_PRI(unit, 7, 3))) |
                            OC_IPC_PRI(unit, pri, sub);

    /* Enable interrupt. */

    SET_REG(OCIEC, OC_INT_FLAGS(unit, OCxIE));
}

/* Enable a unit. */

void oc_on(int unit)
{
    if ((unit < OCMIN) || (unit > OCMAX))
        return;

    SET_REG(OC_REG(unit, OCxCON), 1 << 15);
}



/* Timer configuration. */

void timer_init(int timer, uint8_t prescale, uint16_t limit)
{
    /* NOTE: Should convert from the real prescale value. */

    REG(TIMER_REG(timer, TxCON)) = (prescale & 0b111) << 4;
    REG(TIMER_REG(timer, TMRx)) = 0;
    REG(TIMER_REG(timer, PRx)) = limit;
}

/* Configure interrupts caused by the timer. */

void timer_init_interrupt(int timer, uint8_t pri, uint8_t sub)
{
    if ((timer < TIMERMIN) || (timer > TIMERMAX))
        return;

    /* Disable interrupt and clear interrupt flag. */

    CLR_REG(TIMERIEC, TIMER_INT_FLAGS(timer, TxIE));
    CLR_REG(TIMERIFS, TIMER_INT_FLAGS(timer, TxIF));

    /* Set interrupt priorities. */

    REG(TIMER_IPC_REG(timer)) = (REG(TIMER_IPC_REG(timer)) &
                                 ~(TIMER_IPC_PRI(timer, 7, 3))) |
                                TIMER_IPC_PRI(timer, pri, sub);

    /* Enable interrupt. */

    SET_REG(TIMERIEC, TIMER_INT_FLAGS(timer, TxIE));
}

/* Enable a timer. */

void timer_on(int timer)
{
    if ((timer < TIMERMIN) || (timer > TIMERMAX))
        return;

    SET_REG(TIMER_REG(timer, TxCON), 1 << 15);
}



/* UART configuration. */

void uart_init(int uart, uint32_t baudrate)
{
    /* NOTE: Configured in the initial payload. */

    uint32_t FPB = 24000000;

    if ((uart < UARTMIN) || (uart > UARTMAX))
        return;

    /* Disable the UART (ON). */

    CLR_REG(UART_REG(uart, UxMODE), 1 << 15);

    /* Set the baud rate. For example:

       UxBRG<15:0> = BRG
                   = (FPB / (16 * baudrate)) - 1
                   = (24000000 / (16 * 115200)) - 1
                   = 12
    */

    REG(UART_REG(uart, UxBRG)) = (FPB / (16 * baudrate)) - 1;
}

/* Configure interrupts caused by the UART. */

void uart_init_interrupt(int uart, uint8_t conditions,
                         uint8_t pri, uint8_t sub)
{
    if ((uart < UARTMIN) || (uart > UARTMAX))
        return;

    /* Disable interrupts and clear interrupt flags. */

    CLR_REG(UARTIEC, UART_INT_FLAGS(uart, UxTIE | UxRIE | UxEIE));
    CLR_REG(UARTIFS, UART_INT_FLAGS(uart, UxTIF | UxRIF | UxEIF));

    /* Set priorities: UxIP = pri; UxIS = sub */

    REG(UART_IPC_REG(uart)) = (REG(UART_IPC_REG(uart)) &
                               ~UART_IPC_PRI(uart, 7, 3)) |
                              UART_IPC_PRI(uart, pri, sub);

    /* Enable interrupts. */

    SET_REG(UARTIEC, UART_INT_FLAGS(uart, conditions));
}

/* Enable a UART. */

void uart_on(int uart)
{
    if ((uart < UARTMIN) || (uart > UARTMAX))
        return;

    /* Enable receive (URXEN) and transmit (UTXEN). */

    SET_REG(UART_REG(uart, UxSTA), (1 << 12) | (1 << 10));

    /* Start UART. */

    SET_REG(UART_REG(uart, UxMODE), 1 << 15);
}



/* Utility functions. */

/* Return encoded interrupt priorities. */

static uint8_t PRI(uint8_t pri, uint8_t sub)
{
    return ((pri & 0b111) << 2) | (sub & 0b11);
}

/* Return the DMA interrupt flags for combining with a register. */

int DMA_INT_FLAGS(int channel, uint8_t flags)
{
    return (flags & 0b1) << (DMAINTBASE + (channel - DCHMIN));
}

/* Return encoded DMA interrupt priorities for combining with a register. */

uint32_t DMA_IPC_PRI(int channel, uint8_t pri, uint8_t sub)
{
    return PRI(pri, sub) << (DCHIPCBASE + (channel - DCHMIN) * DCHIPCSTEP);
}

/* Return encoded output compare interrupt priorities for combining with a register. */

uint32_t OC_IPC_PRI(int unit, uint8_t pri, uint8_t sub)
{
    (void) unit;
    return PRI(pri, sub) << OCIPCBASE;
}

/* Return the output compare interrupt priorities register. */

uint32_t OC_IPC_REG(int unit)
{
    switch (unit)
    {
        case 1: return OC1IPC;
        case 2: return OC2IPC;
        case 3: return OC3IPC;
        case 4: return OC4IPC;
        case 5: return OC5IPC;
        default: return 0; /* should not occur */
    }
}

/* Return the output compare interrupt flags for combining with a register. */

int OC_INT_FLAGS(int unit, uint8_t flags)
{
    return (flags & 0b1) << (OCINTBASE + (unit - OCMIN) * OCINTSTEP);
}

/* Return encoded timer interrupt priorities for combining with a register. */

uint32_t TIMER_IPC_PRI(int timer, uint8_t pri, uint8_t sub)
{
    (void) timer;
    return PRI(pri, sub) << TIMERIPCBASE;
}

/* Return the timer interrupt priorities register. */

uint32_t TIMER_IPC_REG(int timer)
{
    switch (timer)
    {
        case 1: return TIMER1IPC;
        case 2: return TIMER2IPC;
        case 3: return TIMER3IPC;
        case 4: return TIMER4IPC;
        case 5: return TIMER5IPC;
        default: return 0; /* should not occur */
    }
}

/* Return the timer interrupt flags for combining with a register. */

int TIMER_INT_FLAGS(int timer, uint8_t flags)
{
    return (flags & 0b1) << (TIMERINTBASE + (timer - TIMERMIN) * TIMERINTSTEP);
}

/* Return encoded UART interrupt priorities for combining with a register. */

uint32_t UART_IPC_PRI(int uart, uint8_t pri, uint8_t sub)
{
    return PRI(pri, sub) << (uart == 1 ? UART1IPCBASE : UART2IPCBASE);
}

/* Return the UART interrupt priorities register. */

uint32_t UART_IPC_REG(int uart)
{
    return uart == 1 ? UART1IPC : UART2IPC;
}

/* Return the UART interrupt flags for combining with a register. */

int UART_INT_FLAGS(int uart, uint8_t flags)
{
    return (flags & 0b111) << (UARTINTBASE + (uart - UARTMIN) * UARTINTSTEP);
}
